@@ -1,6 +1,6 @@
 use crate::{
     tree,
-    tree::{visit::Change, TreeInfo, TreeInfoPair},
+    tree::{visit::Change, TreeInfoPair},
 };
 use git_hash::{oid, ObjectId};
 use git_object::immutable;
@@ -13,7 +13,7 @@ quick_error! {
     #[allow(missing_docs)]
     pub enum Error {
         NotFound { oid: ObjectId } {
-            display("The object {} referenced by the tree was not found in the database", oid)
+            display("The object {} referenced by the tree or the tree itself was not found in the database", oid)
         }
         Cancelled {
             display("The delegate cancelled the operation")
@@ -50,56 +50,50 @@ impl<'a> tree::Changes<'a> {
     ///
     /// [git_cmp_c]: https://github.com/git/git/blob/311531c9de557d25ac087c1637818bd2aad6eb3a/tree-diff.c#L49:L65
     /// [git_cmp_rs]: https://github.com/Byron/gitoxide/blob/a4d5f99c8dc99bf814790928a3bf9649cd99486b/git-object/src/mutable/tree.rs#L52-L55
-    pub fn needed_to_obtain<LocateFn, R, StateMut>(
+    pub fn needed_to_obtain<FindFn, R, StateMut>(
         mut self,
-        other: immutable::TreeIter<'a>,
+        other: immutable::TreeIter<'_>,
         mut state: StateMut,
-        mut locate: LocateFn,
+        mut find: FindFn,
         delegate: &mut R,
     ) -> Result<(), Error>
     where
-        LocateFn: for<'b> FnMut(&oid, &'b mut Vec<u8>) -> Option<immutable::tree::TreeIter<'b>>,
+        FindFn: for<'b> FnMut(&oid, &'b mut Vec<u8>) -> Option<immutable::tree::TreeIter<'b>>,
         R: tree::Visit,
-        StateMut: BorrowMut<tree::State<R::PathId>>,
+        StateMut: BorrowMut<tree::State>,
     {
         let state = state.borrow_mut();
         state.clear();
         let mut lhs_entries = peekable(self.0.take().unwrap_or_default());
         let mut rhs_entries = peekable(other);
-        let mut avoid_popping_path: Option<()> = None;
+        let mut pop_path = false;
 
         loop {
-            if avoid_popping_path.take().is_none() {
+            if pop_path {
                 delegate.pop_path_component();
             }
+            pop_path = true;
+
             match (lhs_entries.next(), rhs_entries.next()) {
                 (None, None) => {
                     match state.trees.pop_front() {
                         Some((None, Some(rhs))) => {
-                            delegate.set_current_path(rhs.parent_path_id.clone());
-                            rhs_entries = peekable(
-                                locate(&rhs.tree_id, &mut state.buf2).ok_or(Error::NotFound { oid: rhs.tree_id })?,
-                            );
+                            delegate.pop_front_tracked_path_and_set_current();
+                            rhs_entries = peekable(find(&rhs, &mut state.buf2).ok_or(Error::NotFound { oid: rhs })?);
                         }
                         Some((Some(lhs), Some(rhs))) => {
-                            delegate.set_current_path(lhs.parent_path_id.clone());
-                            lhs_entries = peekable(
-                                locate(&lhs.tree_id, &mut state.buf1).ok_or(Error::NotFound { oid: lhs.tree_id })?,
-                            );
-                            rhs_entries = peekable(
-                                locate(&rhs.tree_id, &mut state.buf2).ok_or(Error::NotFound { oid: rhs.tree_id })?,
-                            );
+                            delegate.pop_front_tracked_path_and_set_current();
+                            lhs_entries = peekable(find(&lhs, &mut state.buf1).ok_or(Error::NotFound { oid: lhs })?);
+                            rhs_entries = peekable(find(&rhs, &mut state.buf2).ok_or(Error::NotFound { oid: rhs })?);
                         }
                         Some((Some(lhs), None)) => {
-                            delegate.set_current_path(lhs.parent_path_id.clone());
-                            lhs_entries = peekable(
-                                locate(&lhs.tree_id, &mut state.buf1).ok_or(Error::NotFound { oid: lhs.tree_id })?,
-                            );
+                            delegate.pop_front_tracked_path_and_set_current();
+                            lhs_entries = peekable(find(&lhs, &mut state.buf1).ok_or(Error::NotFound { oid: lhs })?);
                         }
                         Some((None, None)) => unreachable!("BUG: it makes no sense to fill the stack with empties"),
                         None => return Ok(()),
                     };
-                    avoid_popping_path = Some(());
+                    pop_path = false;
                 }
                 (Some(lhs), Some(rhs)) => {
                     use std::cmp::Ordering::*;
@@ -125,7 +119,7 @@ impl<'a> tree::Changes<'a> {
 
 fn delete_entry_schedule_recursion<R: tree::Visit>(
     entry: immutable::tree::Entry<'_>,
-    queue: &mut VecDeque<TreeInfoPair<R::PathId>>,
+    queue: &mut VecDeque<TreeInfoPair>,
     delegate: &mut R,
 ) -> Result<(), Error> {
     delegate.push_path_component(entry.filename);
@@ -140,21 +134,15 @@ fn delete_entry_schedule_recursion<R: tree::Visit>(
     }
     if entry.mode.is_tree() {
         delegate.pop_path_component();
-        let path_id = delegate.push_tracked_path_component(entry.filename);
-        queue.push_back((
-            Some(TreeInfo {
-                tree_id: entry.oid.to_owned(),
-                parent_path_id: path_id,
-            }),
-            None,
-        ));
+        delegate.push_back_tracked_path_component(entry.filename);
+        queue.push_back((Some(entry.oid.to_owned()), None));
     }
     Ok(())
 }
 
 fn add_entry_schedule_recursion<R: tree::Visit>(
     entry: immutable::tree::Entry<'_>,
-    queue: &mut VecDeque<TreeInfoPair<R::PathId>>,
+    queue: &mut VecDeque<TreeInfoPair>,
     delegate: &mut R,
 ) -> Result<(), Error> {
     delegate.push_path_component(entry.filename);
@@ -169,14 +157,8 @@ fn add_entry_schedule_recursion<R: tree::Visit>(
     }
     if entry.mode.is_tree() {
         delegate.pop_path_component();
-        let path_id = delegate.push_tracked_path_component(entry.filename);
-        queue.push_back((
-            None,
-            Some(TreeInfo {
-                tree_id: entry.oid.to_owned(),
-                parent_path_id: path_id,
-            }),
-        ))
+        delegate.push_back_tracked_path_component(entry.filename);
+        queue.push_back((None, Some(entry.oid.to_owned())))
     }
     Ok(())
 }
@@ -184,7 +166,7 @@ fn catchup_rhs_with_lhs<R: tree::Visit>(
     rhs_entries: &mut IteratorType<immutable::TreeIter<'_>>,
     lhs: immutable::tree::Entry<'_>,
     rhs: immutable::tree::Entry<'_>,
-    queue: &mut VecDeque<TreeInfoPair<R::PathId>>,
+    queue: &mut VecDeque<TreeInfoPair>,
     delegate: &mut R,
 ) -> Result<(), Error> {
     use std::cmp::Ordering::*;
@@ -224,7 +206,7 @@ fn catchup_lhs_with_rhs<R: tree::Visit>(
     lhs_entries: &mut IteratorType<immutable::TreeIter<'_>>,
     lhs: immutable::tree::Entry<'_>,
     rhs: immutable::tree::Entry<'_>,
-    queue: &mut VecDeque<TreeInfoPair<R::PathId>>,
+    queue: &mut VecDeque<TreeInfoPair>,
     delegate: &mut R,
 ) -> Result<(), Error> {
     use std::cmp::Ordering::*;
@@ -263,13 +245,13 @@ fn catchup_lhs_with_rhs<R: tree::Visit>(
 fn handle_lhs_and_rhs_with_equal_filenames<R: tree::Visit>(
     lhs: immutable::tree::Entry<'_>,
     rhs: immutable::tree::Entry<'_>,
-    queue: &mut VecDeque<TreeInfoPair<R::PathId>>,
+    queue: &mut VecDeque<TreeInfoPair>,
     delegate: &mut R,
 ) -> Result<(), Error> {
     use git_object::tree::EntryMode::*;
     match (lhs.mode, rhs.mode) {
         (Tree, Tree) => {
-            let path_id = delegate.push_tracked_path_component(lhs.filename);
+            delegate.push_back_tracked_path_component(lhs.filename);
             if lhs.oid != rhs.oid
                 && delegate
                     .visit(Change::Modification {
@@ -282,19 +264,10 @@ fn handle_lhs_and_rhs_with_equal_filenames<R: tree::Visit>(
             {
                 return Err(Error::Cancelled);
             }
-            queue.push_back((
-                Some(TreeInfo {
-                    tree_id: lhs.oid.to_owned(),
-                    parent_path_id: path_id.clone(),
-                }),
-                Some(TreeInfo {
-                    tree_id: rhs.oid.to_owned(),
-                    parent_path_id: path_id,
-                }),
-            ));
+            queue.push_back((Some(lhs.oid.to_owned()), Some(rhs.oid.to_owned())));
         }
         (lhs_mode, Tree) if lhs_mode.is_no_tree() => {
-            let path_id = delegate.push_tracked_path_component(lhs.filename);
+            delegate.push_back_tracked_path_component(lhs.filename);
             if delegate
                 .visit(Change::Deletion {
                     entry_mode: lhs.mode,
@@ -313,16 +286,10 @@ fn handle_lhs_and_rhs_with_equal_filenames<R: tree::Visit>(
             {
                 return Err(Error::Cancelled);
             };
-            queue.push_back((
-                None,
-                Some(TreeInfo {
-                    tree_id: rhs.oid.to_owned(),
-                    parent_path_id: path_id,
-                }),
-            ));
+            queue.push_back((None, Some(rhs.oid.to_owned())));
         }
         (Tree, rhs_mode) if rhs_mode.is_no_tree() => {
-            let path_id = delegate.push_tracked_path_component(lhs.filename);
+            delegate.push_back_tracked_path_component(lhs.filename);
             if delegate
                 .visit(Change::Deletion {
                     entry_mode: lhs.mode,
@@ -341,13 +308,7 @@ fn handle_lhs_and_rhs_with_equal_filenames<R: tree::Visit>(
             {
                 return Err(Error::Cancelled);
             };
-            queue.push_back((
-                Some(TreeInfo {
-                    tree_id: lhs.oid.to_owned(),
-                    parent_path_id: path_id,
-                }),
-                None,
-            ));
+            queue.push_back((Some(lhs.oid.to_owned()), None));
         }
         (lhs_non_tree, rhs_non_tree) => {
             delegate.push_path_component(lhs.filename);
